unit U_types;

interface

uses u_adt;

   const  { zunchst die verschiedenen Mengen von Zeichen: }

          Buchstaben     : set of char = ['a'..'z','A'..'Z','_'];
          ziffern        : set of char = ['0'..'9'];
          kontrollzeichen: set of char = [#10,#13,#32];

          { und schlielich die festen Max-Konstanten: }

          maxSymbole          = 100; { maximal 100 verschiedene Symbole              }
          maxBezeichnerLaenge =  20; { Bezeichnerlnge ist auf 20 Zeichen beschrnkt }

   type
         { Der Typ TSymbol beschreibt die unterschiedlichen tokens ! }
          TSymbol       = (prog,          start,           ende,            punkt,   klammerauf, klammerzu,
                           zuweisungsop,  punktoperator,   strichoperator,  zahl,    name,
                           semikolon,     lies,            schreib,         wenn,    dann,       sonst,
                           vergleichsop,  wiederhole,      bis,             solange, tue,        und,
                           oder,          nicht_zulaessig, leer);

          TSymbolindex   = 0..maxSymbole;

          TSymbolString  = string [maxBezeichnerLaenge];

          TSymbolEintrag =  record
                                 Zeichenkette  : TSymbolString ;
                                 Symbol        : TSymbol       ;
                            end;

          TSymbolTabelle = class
            private
            Items: array [1..maxSymbole] of TSymbolEintrag ;

            public
            constructor create;
            function SymbolString ( i: TSymbolindex  ) : String;
            function Symbol       ( i: TSymbolindex  ) : TSymbol ;
            function SymbolKuerzel( symbol : TSymbol ) : String ;

            function SymbolIndex  (Zeichenkette: TSymbolstring): TSymbolindex;
          end;

          TSymbolelement = class(TElement)
            Inhalt: TSymbolindex;
            constructor create(i: TSymbolindex);
          end;

          TTokenliste    = class(TLinList)       { ber dem Typ TSymbolindex }
            private
            Symboltabelle: TSymboltabelle;

            public
            constructor create(ST: TSymboltabelle);
            function NextTokenIndex: TSymbolIndex;
            function NextToken: TSymbol;
          end;


              { Man htte zwar auch den ADT SCHLANGE verwenden knnen, dann  }
              { wrde aber bei Abarbeitung der tokenliste durch den Parser   }
              { die Schlange nicht mehr existieren ! [...nach dem Entfernen  }
              { des Schlangenkopfes wird der zugehrige Speicherraum durch   }
              { 'dispose' wieder freigegeben.] Z.B. mchte man fr sptere   }
              { Ausgaben die tokenliste noch zur Verfgung haben !           }

implementation

{ ------------------     Routinen fr die Symboltabelle    ---------------------------------------- }

constructor TSymboltabelle.create;
var   i : integer ;
begin
   for i := 1  to  maxSymbole
   do with Items[i]
   do begin
        Zeichenkette  := ''   ;
        Symbol  := leer ;
      end;
   Items[ 1].symbol := klammerauf     ; Items[ 1].Zeichenkette:= '('      ;
   Items[ 2].symbol := klammerzu      ; Items[ 2].Zeichenkette:= ')'      ;
   Items[ 3].symbol := punkt          ; Items[ 3].Zeichenkette:= '.'      ;
   Items[ 4].symbol := zuweisungsop   ; Items[ 4].Zeichenkette:= ':='     ;
   Items[ 5].symbol := prog           ; Items[ 5].Zeichenkette:= 'program';
   Items[ 6].symbol := start          ; Items[ 6].Zeichenkette:= 'begin'  ;
   Items[ 7].symbol := ende           ; Items[ 7].Zeichenkette:= 'end'    ;
   Items[ 8].symbol := strichoperator ; Items[ 8].Zeichenkette:= '+'      ;
   Items[ 9].symbol := strichoperator ; Items[ 9].Zeichenkette:= '-'      ;
   Items[10].symbol := punktoperator  ; Items[10].Zeichenkette:= '*'      ;
   Items[11].symbol := punktoperator  ; Items[11].Zeichenkette:= '/'      ;

   Items[12].symbol := semikolon      ; Items[12].Zeichenkette:= ';'      ;
   Items[13].symbol := lies           ; Items[13].Zeichenkette:= 'readln' ;
   Items[14].symbol := schreib        ; Items[14].Zeichenkette:= 'writeln';

   Items[15].symbol := wenn           ; Items[15].Zeichenkette:= 'if'     ;
   Items[16].symbol := dann           ; Items[16].Zeichenkette:= 'then'   ;
   Items[17].symbol := sonst          ; Items[17].Zeichenkette:= 'else'   ;

   Items[18].symbol := vergleichsop   ; Items[18].Zeichenkette:= '='      ;
   Items[19].symbol := vergleichsop   ; Items[19].Zeichenkette:= '>'      ;
   Items[20].symbol := vergleichsop   ; Items[20].Zeichenkette:= '>='     ;
   Items[21].symbol := vergleichsop   ; Items[21].Zeichenkette:= '<='     ;
   Items[22].symbol := vergleichsop   ; Items[22].Zeichenkette:= '<'      ;
   Items[23].symbol := vergleichsop   ; Items[23].Zeichenkette:= '<>'     ;

   Items[24].symbol := solange        ; Items[24].Zeichenkette:= 'while'  ;
   Items[25].symbol := tue            ; Items[25].Zeichenkette:= 'do'     ;
   Items[26].symbol := wiederhole     ; Items[26].Zeichenkette:= 'repeat' ;
   Items[27].symbol := bis            ; Items[27].Zeichenkette:= 'until'  ;

   Items[28].symbol := und            ; Items[28].Zeichenkette:= 'and'    ;
   Items[29].symbol := oder           ; Items[29].Zeichenkette:= 'or'     ;

   Items[30].symbol := punktoperator  ; Items[30].Zeichenkette:= 'mod'    ;

   Items[31].symbol := nicht_zulaessig; Items[31].Zeichenkette:= '??'     ;
end;


function TSymboltabelle.SymbolIndex(Zeichenkette: TSymbolstring): TSymbolindex;
var i: TSymbolIndex;
begin
   i := 0;
   repeat
     i := succ(i);
   until (Items[i].symbol = leer) or (Items[i].Zeichenkette = Zeichenkette)
                                  or (i = maxSymbole-1) ;
      { i=max_Symbole bleibt dem token 'leer' bzw. 'nn' vorbehalten ! }

   if (Items[i].Zeichenkette <> ZeichenKette) and (i = maxSymbole)
   then SymbolIndex:= 0             { ---- Fall: Symboltabelle zu klein ausgelegt }

   else begin
          if Items[i].Symbol = leer { ---- Fall: Bezeichner oder Zahl neu anzulegen   }
          then begin
                 Items[i].Zeichenkette := Zeichenkette;
                 if Zeichenkette[1] in Buchstaben
                 then Items[i].Symbol:= name
                 else Items[i].Symbol:= zahl;
               end;
          SymbolIndex:= i
        end
 end;

  function TSymboltabelle.SymbolString(i: TSymbolIndex): String;
  begin
    SymbolString:= Items[i].Zeichenkette
  end;

  function TSymboltabelle.Symbol(i: TSymbolIndex): TSymbol;
  begin
    Symbol:= Items[i].symbol
  end;

  function TSymboltabelle.SymbolKuerzel( symbol: TSymbol ): String ;    {eine Klassifizierung}
  var s: String;
  begin
    case symbol of name                         : s:= 'Id' ;
                   zahl                         : s:= 'Zi' ;
                   punktoperator,strichoperator,
                   vergleichsop, zuweisungsop   : s:= 'Op' ;
                   punkt                        : s:= 'Pk' ;
                   nicht_zulaessig              : s:= '??' ;
                   leer                         : s:= 'nn' ;
       {Sonderzeichen}
                   klammerauf, klammerzu,
                   semikolon                    : s:= 'Sz' ;
       {Schlsselwort}
                   prog, start, ende,
                   lies, schreib,
                   wenn, dann, sonst,
                   wiederhole, bis,
                   solange, tue, und, oder      : s:= 'Sw' ;
    end ;
    SymbolKuerzel:= s
  end ;


{ ------------------     Routinen fr die Tokenliste    ------------------------------------------- }

constructor TSymbolElement.create(i: TSymbolindex);
begin
  inherited create;
  Inhalt:= i;
end;

constructor TTokenliste.create(ST: TSymboltabelle);
begin
  inherited create;
  Symboltabelle:= ST;
end;

function TTokenliste.NextTokenIndex: TSymbolIndex;
var nr: integer;
begin
  begin
    if not leer                            {die tokenliste wird durch 'leer' }
    then begin                             {vom Scanner abgeschlossen !      }
           nr:= (Aktuell as TSymbolElement).Inhalt;
           Vor;                            {kein Effekt bei erreichtem ende !}
         end
    else nr := maxSymbole;                 {d.h. leer ! }
    NextTokenIndex:= nr;
  end;

end;

function TTokenliste.NextToken: TSymbol;
var nr: integer;
begin
  nr:= NextTokenIndex;
  NextToken:= Symboltabelle.Symbol(nr);
end;

end.
